//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  fbase.cpp
//! \brief Collection of basic and arithmetic functions.


#include <iostream>
#include <map>
#include <vector>
#include <cstdlib>
#include <math.h>
#include <float.h>
#include <gsl/gsl_sf.h>

#include <../trivia/rng.hpp>
#include <kww.h>

#include "defs.hpp"
#include "func.hpp"
#include "fregistry.hpp"

#include <cerf.h> // include after boost::shared_ptr, otherwise use of 'I' in template clashes

const double PI    = M_PI;    //4*atan(1.);
const double TWOPI = 2*M_PI; //8*atan(1.);

static triv::RandomNumberGenerator *rng = NULL;


//**************************************************************************************************
//*  Functions of one argument
//**************************************************************************************************

template<class A>
void func_pop(A v) {;}

double func_to_d(int v) { return v; }
double func_to_d(double v) { return v; }
double func_to_d(string s)
{
    double ret;
    if ( triv::any2dbl( s, &ret ) )
        return ret;
    throw "Cannot convert "+s+" to floating-point number";
}

template<class A>
int func_not(A v) { return !v; }

template<class A>
A func_neg(A v) { return -v; }
void func_neg(double& r, double& dr, double v, double dv) { r=-v; dr=dv; }

double func_ln(double v) { return v>0 ? log(v) : 0; }
void func_ln(double& r, double& dr, double v, double dv)
{ if(v>0) { r=log(v); dr=dv/v; } else { r=0; dr=0; } }

double func_lg(double v) { return v>0 ? log10(v) : 0; }
void func_lg(double& r, double& dr, double v, double dv)
{ if(v>0) { r=log10(v); dr=dv/v/log(10); } else { r=0; dr=0; } }

double func_sqrt(double v) { return v>=0 ? sqrt(v) : 0; }
void func_sqrt(double& r, double& dr, double v, double dv)
{ if(v>0) { r=sqrt(v); dr=dv/2/r; } else { r=0; dr=0; } }

double func_abs(double v) { return fabs(v); }
int    func_abs(int v)    { return abs(v); }
void   func_abs(double& r,double& dr, double v,double dv) { r=fabs(v); dr=dv; }

template<class A>
A func_sign(A v) {
    if      ( v<0 )
        return -1;
    else if ( v>0 )
        return 1;
    else
        return 0;
}
void func_sign(double& r,double& dr, double v,double dv) { r=func_sign(v); dr=dv; }

double func_exp(double v) { return exp(v); }
void func_exp(double& r, double& dr, double v, double dv) { r=exp(v); dr=dv*r; }

double func_sin(double v) { return sin(v); }
void func_sin(double& r, double& dr, double v, double dv) { r=sin(v); dr=dv*fabs(cos(v)); }
double func_cos(double v) { return cos(v); }
void func_cos(double& r, double& dr, double v, double dv) { r=cos(v); dr=dv*fabs(sin(v)); }
double func_tan(double v) { return tan(v); }
double func_cot(double v) { double t=tan(v); return t==0?0:1/t;}

double func_sind(double v) { return sin(v*TWOPI/360.); }
double func_cosd(double v) { return cos(v*TWOPI/360.); }
double func_tand(double v) { return tan(v*TWOPI/360.); }
double func_cotd(double v) { double t=tan(v*TWOPI/360.);  return t==0?0:1/t; }

double func_asin(double v) { return (-1<=v && v<=1) ? asin(v) : 0; }
double func_acos(double v) { return (-1<=v && v<=1) ? acos(v) : 0; }
double func_atan(double v) { return atan(v); }
double func_acot(double v) { return v!=0 ? atan(1/v) : TWOPI/4; }

double func_asind(double v) { return (-1<=v && v<=1) ? 360./TWOPI*asin(v) : 0; }
double func_acosd(double v) { return (-1<=v && v<=1) ? 360./TWOPI*acos(v) : 0; }
double func_atand(double v) { return 360./TWOPI*atan(v); }
double func_acotd(double v) { return v!=0 ? 360./TWOPI*atan(1/v) : 90; }

double func_sinh(double v) { return sinh(v); }
double func_cosh(double v) { return cosh(v); }
double func_tanh(double v) { return tanh(v); }
double func_coth(double v) { double t=tanh(v); return t==0?0:1/t;}

double func_j0(double v) { return v==0 ? 1 : sin(v)/v; }
double func_j1(double v) { return v==0 ? 1 : (sin(v)/v-cos(v))/v; }

double func_fac(double v) { return v>-.5 ? gsl_sf_gamma(round(v+1)) : 0; }
double func_cata(double v) { // Catalan number ?
    if ( v<=-.5 )
        return 0;
    double x = round(v);
    return exp( gsl_sf_lngamma( 2*x+1 ) - gsl_sf_lngamma( x+1 ) -
                gsl_sf_lngamma( x+2 ) ); }

double func_gamma(double v) { return v>0 ? gsl_sf_gamma(v) : 0; }
double func_lgamma(double v) { return v>0 ? gsl_sf_lngamma(v) : 0; }
// could be extended to non-integer negative arguments...

double func_debye1(double v) { return v>=0 ? gsl_sf_debye_1(v) : 0; }
double func_debye3(double v) { return v>=0 ? gsl_sf_debye_3(v) : 0; }
double func_debye_msd(double v) { // mean squared displacement
                                  // without prefactors
                                  // without zero temperature oscillations
    return v>1e-2 ? 2*v*gsl_sf_debye_1(1/v) : 0; }
double func_debye_ui(double v) { // internal energy
    return v>=0 ? 3*v*gsl_sf_debye_3(1/v) : 0; }
double func_debye_cv(double v) { // heat capacity
    double eps=1e-12;
    return v>=eps ? (func_debye_ui(v+eps)-func_debye_ui(v-eps))/(2*eps) : 0; }

double func_erfP(double v) { return 1-gsl_sf_erf_Q(v); }
double func_erfQ(double v) { return gsl_sf_erf_Q(v); }
double func_erf(double v) { return gsl_sf_erf(v); }
double func_erfc(double v) { return gsl_sf_erfc(v); }
double func_erfi(double v) { return erfi(v); }

double func_dawson(double v) { return gsl_sf_dawson(v); }

double func_sinc(double v) { return v==0 ? 1 : sin(v)/v; }

int func_ceil(double v)  { return (int)(ceil(v)+0.5); }
int func_floor(double v) { return (int)(floor(v)+0.5); }
int func_nint(double v)  { return (int)(round(v)+0.5); }

// Ad hoc for uses in physics:

double func_lndiehl( double a ) {
    // ln( \int dw gauss(w,gausswid) cauchy(w,cauchywid) / gauss(0,gausswid) )
    // argumnt a = cauchywid/gausswid
    return SQR(a)/2 + gsl_sf_log_erfc( a/sqrt(2) );
}

double func_diehl( double a ) {
    // \int dw gauss(w,gausswid) cauchy(w,cauchywid) / gauss(0,gausswid)
    // argumnt a = cauchywid/gausswid
    return exp( func_lndiehl(a) );
}

int func_exit( int a ) { exit(a); return 0; }

int func_throw_if( int a, string s ) { if(a) throw s; return 0; }
int func_throw_unless( int a, string s ) { if(!a) throw s; return 0; }

//**************************************************************************************************
//*  Functions of two arguments
//**************************************************************************************************

//double func_pm(double v, double a) { return v; }
void func_pm(double& r, double& dr, double v, double a){ r=v; dr=a; }

template<class A>
A func_add(A v, A a) { return v+a; }
void func_add(double& r, double& dr, double v, double dv, double a, double da)
{ r=v+a; dr=sqrt(dv*dv+da*da); }

template<class A>
A func_sub(A v, A a) { return v-a; }
void func_sub(double& r, double& dr, double v, double dv, double a, double da)
{ r=v-a; dr=sqrt(dv*dv+da*da); }

template<class A>
A func_mul(A v, A a) { return v*a; }
void func_mul(double& r, double& dr, double v, double dv, double a, double da)
{ r=v*a; dr=sqrt(SQR(v*da)+SQR(a*dv)); }

double func_div(double v, double a) { return a!=0 ? v/a : 0; }
void func_div(double& r, double& dr, double v, double dv, double a, double da)
{ if(a==0) { r=0; dr=0; } else { r=v/a; dr=sqrt(SQR(dv/a)+SQR(v*da/a/a)); } }

int func_idiv(int v, int a) { return a!=0 ? v/a : 0; }
int func_idiv(double v, int a) { return a!=0 ? (int)(ceil(v)+0.5)/a : 0; }

int func_mod(int v, int a) { return v%a; }
double func_mod(double v, double a) { return fmod(v,a); }

double func_pow(double v, double a)
{ return ( v>0 || (v==0&&a>0) || a==(int)a ) ? pow(v,a) : 0; }
void func_pow(double& r, double& dr, double v, double dv, double a, double da)
{
    if ( v>0 || (v==0&&a>0) || a==(int)a ) {
        r = pow(v,a); dr = sqrt( SQR(a*r/v) + SQR(log(v)*a) );
    } else {
        r = 0; dr = 0;
    }
}

template<class A>
int func_eq (A v, A a) { return v == a; }
template<class A>
int func_ne (A v, A a) { return v != a; }
template<class A>
int func_gt (A v, A a) { return v > a; }
template<class A>
int func_ge (A v, A a) { return v >= a; }
template<class A>
int func_lt (A v, A a) { return v < a; }
template<class A>
int func_le (A v, A a) { return v <= a; }
template<class A>
int func_and(A v, A a) { return v && a; }
template<class A>
int func_or (A v, A a) { return v || a; }
template<class A>
int func_xor(A v, A a) { return (bool) v ^ (bool) a; }

double func_over(double v, double a) {
    return  ( 0<=v && 0<=a && a<=v && v <= GSL_SF_GAMMA_XMAX ) ?
        gsl_sf_gamma(v+1)/gsl_sf_gamma(a+1)/gsl_sf_gamma(v-a+1) :
        0; }

template<class A>
A func_min (A v, A a) { return v<a ? v : a; }
template<class A>
A func_max (A v, A a) { return v>a ? v : a; }

double func_gauss (double x, double s) {
    return s==0 ? 0 : exp(-x*x/(2*s*s))/sqrt(TWOPI*s*s); }
double func_gaussnn (double x, double s) { // Gauss not normalized
    return s==0 ? 0 : exp(-x*x/(2*s*s)); }
double func_cauchy (double x, double s) {
    return s==0 ? 0 : fabs(s)/PI/(x*x+s*s); }

double func_ran (double v, double a) {
    // a random number between v and a :
    if (!rng)
        rng = new triv::RandomNumberGenerator();
    return rng->uniform( func_min(v,a), func_max(v,a) ); }

// Havriliak-Negami: unchecked formulae from disreputable source:

double aux_phi_havneg (double v, double a){
    return atan( pow(v,a)*sin(M_PI*a/2)/(1+pow(v,a)*cos(M_PI*a/2)) ); }

double func_re_havneg (double v, double a, double g) {
    return pow( 1 + 2*pow(v,a)*cos(M_PI*a/2)+pow(v,2*a), -g/2 ) *
        cos( g*aux_phi_havneg( v, a ) ); }

double func_im_havneg (double v, double a, double g) {
    return pow( 1 + 2*pow(v,a)*cos(M_PI*a/2)+pow(v,2*a), -g/2 ) *
        sin( g*aux_phi_havneg( v, a ) ); }


//**************************************************************************************************
//*  Functions of three arguments
//**************************************************************************************************

int func_cond(int v, int a, int b) { return v ? a : b; }
template<class A, class B>
double func_cond(int v, A a, B b) { return v ? a : b; }
void func_cond(double& r, double& dr, double v, double dv, double a, double da, double b, double db)
{ if(v) { r=a; dr=da; } else { r=b; dr=db; } }

double func_q4w( double hbarw, double Ei, double twotheta) {
    // calculate Q in Angstroem-1 when hbar*w and E_i are given in meV
    if( hbarw>Ei ) return -1;
    return sqrt( (1/2.07212349) *
                 ( 2*Ei-hbarw-2*sqrt(Ei*(Ei-hbarw))*cos(twotheta*TWOPI/360) ) );
}

double func_cauchy2 (double x, double p, double w) {
    return w==0 ? 0 : w/PI*(1.0/(SQR(x-p)+SQR(w))+1.0/(SQR(x+p)+SQR(w)));
}

double func_kwwc( double v, double t, double b ){
    if( b<0.1 || b>2.0 ) return 0;
    return t/PI*kwwc( v*t, b ); }
double func_kwws( double v, double t, double b ){
    if( b<0.1 || b>2.0 ) return 0;
    return t/PI*kwws( v*t, b ); }
double func_kwwp( double v, double t, double b ){
    if( b<0.1 || b>2.0 || t<=0 ) return 0;
    return 1/PI*kwwp( v*t, b ); }

double func_voigt( double v, double s, double g ){
    return (s||g) ? voigt(v,s,g) : 0;
}

double func_zorn( double v, double vavg, double s ) {
    if( s<=0 || s>=1 ) return 0;
    return (1-s)*(v+s*vavg*vavg/(1-s*vavg)); }

double func_zorn_gauss( double q, double u2, double s ) {
    if( s<=0 || s>=1 ) return 0;
    double qmax = 4*PI/6.27; // Si111
    double v = exp(-u2*SQR(q));
    double vavg = (1-exp(-u2*SQR(qmax))) / (u2*SQR(qmax));
    return (1-s)*(v+s*vavg*vavg/(1-s*vavg)); }

double func_rrdm( double wred, double act0, double relwidth ) {
    // Rotational rate distribution model [review: Colmenero 2005]
    //   Gaussian distribution of Barriers, Arrhenius law, rotational jumps.
    // Dimensionless:
    //   wred = w/Omega_0,
    //   act0 = mean{E_A}/(k_B T),
    //   relwidth = stdv{E_A} / mean{E_A};
    //   multiply result by T/Omega_0.
    const int N=101;
    double actsig = act0*relwidth;
    double actrange = 3*actsig;
    if ( act0-actrange <= 0 )
        return 0;
    double da = actrange/((N-1)/2);
    double ret = 0;
    for( int i=0; i<N; ++i ) {
        double a = act0 + da*(i-(N-1)/2);
        ret += exp( -SQR((a-act0)/actsig)/2 ) * func_cauchy( wred, exp(-a) );
    }
    return ret * da / sqrt(2*PI) / actsig;
}

double func_rotdiff( double w, double tau, double qb )
// Rotational diffusion spectrum
{
    double ret = 0;
    if( qb<=0 )
        return 0;
    for( int l=1; l<=12; ++l ){
        ret += (2*l+1) * pow( gsl_sf_bessel_jl( l, qb ), 2 ) *
            l*(l+1)*tau/PI/(SQR(w*tau)+SQR(l*(l+1)));
    }
    return ret;
}

double func_ornuhl( double _w, double _a, double _b )
// Ornstein-Uhlenbeck spectrum FT[exp(-|b|t-a(1-exp(-t))),
//    computed for w->0 by expanding the outer exponential in e^(a e^(-x/w))
//    which turns out to converge rapidly for all w
{
    const double ornuhl_delta=2.2e-16, ornuhl_eps=5.5e-20;
    long double w = _w;
    long double a = _a;
    long double b = fabs(_b);
    if ( a==0 )
        return b/(SQR(b)+SQR(w));
    long double a_log = logl( a );
    long double S=0;      // summed series
    long double h;        // exponent
    long double f;        // Cauchy-Lorentz function
    long double u;        // precomputed common factors
    long double u_next=0; // - next value [initialized to avoid warning]

    // sum the expansion
    for ( int k=0; k<1000; ++k ) {
        u = u_next;
        h = k*a_log - lgammal((long double)(k+1));
        if ( h>DBL_MAX_EXP )
            return -3; // exponent overflow
        if ( h<DBL_MIN_EXP )
            return -4; // exponent underflow
        f = (k+b)/(SQR(k+b)+SQR(w));
        u_next = expl( h ) * f;
        S += u;
        if( k==0 )
            continue;
        // termination criteria
        if ( ornuhl_eps*S+u_next <= ornuhl_delta*S )
            return exp(-a)*S; // reached required precision
        else if ( u_next<DBL_MIN )
            return -7; // underflow
    }
    return -9; // too many terms
}

//**************************************************************************************************
//*  Registration
//**************************************************************************************************

//! Registers basic and arithmetic functions.

void fbase_initialize()
{

#define G GFuncRegistry

    // operators by precedence (which should agree with xax_yacc.ypp):
    G->register_fct_meta ( "+-",  2, "a+-da   [set error]", 1 );
    G->register_fct_e_dd ( "+-", func_pm );
    G->register_fct_meta ( "^",   2, "a^b     [exponentiation]", 2 );
    G->register_fct_d_dd ( "^", func_pow );
    G->register_fct_e_ee ( "^", func_pow );
    G->register_fct_meta ( "!",   1, "!a      [negation]", 3 );
    G->register_fct_i_i  ( "!", func_not );
    G->register_fct_i_d  ( "!", func_not );
    G->register_fct_meta ( "neg", 1, "-a      [negative value]", 4 );
    G->register_fct_i_i  ( "neg", func_neg );
    G->register_fct_d_d  ( "neg", func_neg );
    G->register_fct_e_e  ( "neg", func_neg );
    G->register_fct_meta ( "/",   2, "a/b     [quotient]", 5 );
    G->register_fct_d_dd ( "/", func_div );
    G->register_fct_e_ee ( "/", func_div );
    G->register_fct_meta ( "//",  2, "i//j    [integer quotient]", 5 );
    G->register_fct_i_ii ( "//", func_idiv );
    G->register_fct_i_di ( "//", func_idiv );
    G->register_fct_meta ( "%",   2, "i%j     [remainder]", 5 );
    G->register_fct_i_ii ( "%", func_mod );
    G->register_fct_d_dd ( "%", func_mod );
    G->register_fct_meta ( "*",   2, "a*b     [product]", 6 );
    G->register_fct_i_ii ( "*", func_mul );
    G->register_fct_d_dd ( "*", func_mul );
    G->register_fct_e_ee ( "*", func_mul );
    G->register_fct_meta ( "+",   2, "a+b     [sum]", 7 );
    G->register_fct_i_ii ( "+", func_add );
    G->register_fct_d_dd ( "+", func_add );
    G->register_fct_e_ee ( "+", func_add );
    G->register_fct_meta ( "-",   2, "a-b     [difference]", 7 );
    G->register_fct_i_ii ( "-", func_sub );
    G->register_fct_d_dd ( "-", func_sub );
    G->register_fct_e_ee ( "-", func_sub );
    G->register_fct_meta ( ">",   2, "a>b     [greater]", 8 );
    G->register_fct_i_ii ( ">", func_gt );
    G->register_fct_i_dd ( ">", func_gt );
    G->register_fct_meta ( "<",   2, "a<b     [less]", 8 );
    G->register_fct_i_ii ( "<", func_lt );
    G->register_fct_i_dd ( "<", func_lt );
    G->register_fct_meta ( ">=",  2, "a>=b    [greater or equal]", 8 );
    G->register_fct_i_ii ( ">=", func_ge );
    G->register_fct_i_dd ( ">=", func_ge );
    G->register_fct_meta ( "<=",  2, "a<=b    [less or equal]", 8 );
    G->register_fct_i_ii ( "<=", func_le );
    G->register_fct_i_dd ( "<=", func_le );
    G->register_fct_meta ( "==",  2, "a==b    [equality]", 9 );
    G->register_fct_i_ii ( "==", func_eq );
    G->register_fct_i_dd ( "==", func_eq );
    G->register_fct_meta ( "!=",  2, "a!=b    [unequality]", 9 );
    G->register_fct_i_ii ( "!=", func_ne );
    G->register_fct_i_dd ( "!=", func_ne );
    G->register_fct_meta ( "xor", 2, "a^b     [exclusive or]", 10 );
    G->register_fct_i_ii ( "xor", func_xor );
    G->register_fct_i_dd ( "xor", func_xor );
    G->register_fct_meta ( "&&",  2, "a&&b    [and]", 11 );
    G->register_fct_i_ii ( "&&", func_and );
    G->register_fct_i_dd ( "&&", func_and );
    G->register_fct_meta ( "||",  2, "a||b    [or]", 11 );
    G->register_fct_i_ii ( "||", func_or );
    G->register_fct_i_dd ( "||", func_or );
    G->register_fct_meta ( "?:",  2, "a?b:c   [if a then b else c]", 13 );
    G->register_fct_i_iii( "?:", func_cond );
    G->register_fct_d_iid( "?:", func_cond );
    G->register_fct_d_idi( "?:", func_cond );
    G->register_fct_d_idd( "?:", func_cond );
    G->register_fct_e_eee( "?:", func_cond );


    // f( 1 arg )
    G->register_fct_meta ( "to_d", 1, "(x): convert x to floating-point number" );
    G->register_fct_d_i  ( "to_d", func_to_d );
    G->register_fct_d_d  ( "to_d", func_to_d );
    G->register_fct_d_s  ( "to_d", func_to_d );
    G->register_fct_meta ( "pop", 1, "(x): do nothing" );
    G->register_fct_0_i  ( "pop", func_pop );
    G->register_fct_0_d  ( "pop", func_pop );
    G->register_fct_0_s  ( "pop", func_pop );
    G->register_fct_meta ( "exit", 1, "(x): exit(x)" );
    G->register_fct_i_i  ( "exit", func_exit );
    G->register_fct_meta ( "ln", 1, "(x): natural logarithm of x, or 0 if x<=0" );
    G->register_fct_d_d  ( "ln", func_ln );
    G->register_fct_e_e  ( "ln", func_ln );
    G->register_fct_meta ( "lg", 1, "(x): decadic logarithm of x, or 0 if x<=0" );
    G->register_fct_d_d  ( "lg", func_lg );
    G->register_fct_e_e  ( "lg", func_lg );
    G->register_fct_meta ( "sqrt", 1, "(x): square root of x, or 0 if x<0" );
    G->register_fct_d_d  ( "sqrt", func_sqrt );
    G->register_fct_e_e  ( "sqrt", func_sqrt );
    G->register_fct_meta ( "abs", 1, "(x): absolute value of x" );
    G->register_fct_i_i  ( "abs", func_abs );
    G->register_fct_d_d  ( "abs", func_abs );
    G->register_fct_e_e  ( "abs", func_abs );
    G->register_fct_meta ( "sign", 1, "(x): sign of x" );
    G->register_fct_i_i  ( "sign", func_sign );
    G->register_fct_d_d  ( "sign", func_sign );
    G->register_fct_e_e  ( "sign", func_sign );
    G->register_fct_meta ( "exp", 1, "(x): exponential function of x" );
    G->register_fct_d_d  ( "exp", func_exp );
    G->register_fct_e_e  ( "exp", func_exp );
    G->register_fct_meta ( "sin", 1, "(x): sine of x (where x in radian)" );
    G->register_fct_d_d  ( "sin", func_sin );
    G->register_fct_e_e  ( "sin", func_sin );
    G->register_fct_meta ( "cos", 1, "(x): cosine of x (where x in radian)" );
    G->register_fct_d_d  ( "cos", func_cos );
    G->register_fct_e_e  ( "cos", func_cos );
    G->register_fct_meta ( "tan", 1, "(x): tangent of x (where x in radian)" );
    G->register_fct_d_d  ( "tan", func_tan );
    G->register_fct_meta ( "cot", 1, "(x): cotangent of x (where x in radian)" );
    G->register_fct_d_d  ( "cot", func_cot );
    G->register_fct_meta ( "sind", 1, "(x): sine of x (where x in degrees)" );
    G->register_fct_d_d  ( "sind", func_sind );
    G->register_fct_meta ( "cosd", 1, "(x): cosine of x (where x in degrees)" );
    G->register_fct_d_d  ( "cosd", func_cosd );
    G->register_fct_meta ( "tand", 1, "(x): tangent of x (where x in degrees)" );
    G->register_fct_d_d  ( "tand", func_tand );
    G->register_fct_meta ( "cotd", 1, "(x): cotangent of x (where x in degrees)" );
    G->register_fct_d_d  ( "cotd", func_cotd );
    G->register_fct_meta ( "asin", 1, "(x): arc sine of x (result is in radian; 0 if |x|>1)" );
    G->register_fct_d_d  ( "asin", func_asin );
    G->register_fct_meta ( "acos", 1, "(x): arc cosine of x (result is in radian; 0 if |x|>1)" );
    G->register_fct_d_d  ( "acos", func_acos );
    G->register_fct_meta ( "atan", 1, "(x): arc tangent of x (result is in radian)" );
    G->register_fct_d_d  ( "atan", func_atan );
    G->register_fct_meta ( "acot", 1, "(x): arc cotangent of x (result is in radian)" );
    G->register_fct_d_d  ( "acot", func_acot );
    G->register_fct_meta ( "asind", 1, "(x): arc sine of x (result is in degrees; 0 if |x|>1)" );
    G->register_fct_d_d  ( "asind", func_asind );
    G->register_fct_meta ( "acosd", 1, "(x): arc cosine of x (result is in degrees; 0 if |x|>1)" );
    G->register_fct_d_d  ( "acosd", func_acosd );
    G->register_fct_meta ( "atand", 1, "(x): arc tangent of x (result is in degrees)" );
    G->register_fct_d_d  ( "atand", func_atand );
    G->register_fct_meta ( "acotd", 1, "(x): arc cotangent of x (result is in degrees)" );
    G->register_fct_d_d  ( "acotd", func_acotd );
    G->register_fct_meta ( "sinh", 1, "(x): hyperbolic sine of x" );
    G->register_fct_d_d  ( "sinh", func_sinh );
    G->register_fct_meta ( "cosh", 1, "(x): hyperbolic cosine of x" );
    G->register_fct_d_d  ( "cosh", func_cosh );
    G->register_fct_meta ( "tanh", 1, "(x): hyperbolic tangent of x" );
    G->register_fct_d_d  ( "tanh", func_tanh );
    G->register_fct_meta ( "coth", 1, "(x): hyperbolic tangent of x" );
    G->register_fct_d_d  ( "coth", func_coth );
    G->register_fct_meta ( "j0", 1, "(x): spherical Bessel function x" );
    G->register_fct_d_d  ( "j0", func_j0 );
    G->register_fct_meta ( "j1", 1, "(x): spherical Bessel function x" );
    G->register_fct_d_d  ( "j1", func_j1 );
    G->register_fct_meta ( "fac", 1, "(x): factorial of nearest integer of x" );
    G->register_fct_d_d  ( "fac", func_fac );
    G->register_fct_meta ( "cata", 1, "(x): Catalan number of nearest integer of x" );
    G->register_fct_d_d  ( "cata", func_cata );
    G->register_fct_meta ( "gamma", 1, "(x): gamma function of x (i.e. factorial of x-1)" );
    G->register_fct_d_d  ( "gamma", func_gamma );
    G->register_fct_meta ( "lgamma", 1, "(x): ln of gamma of x" );
    G->register_fct_d_d  ( "lgamma", func_lgamma );
    G->register_fct_meta ( "debye1", 1, "(x): Debye function D1 of x" );
    G->register_fct_d_d  ( "debye1", func_debye1 );
    G->register_fct_meta ( "debye3", 1, "(x): Debye function D3 of x" );
    G->register_fct_d_d  ( "debye3", func_debye3 );
    G->register_fct_meta ( "debyeu2", 1, "(x): Debye mean squared displacement's temperature dependence" );
    G->register_fct_d_d  ( "debyeu2", func_debye_msd );
    G->register_fct_meta ( "debyeui", 1, "(x): Debye internal energy" );
    G->register_fct_d_d  ( "debyeui", func_debye_ui );
    G->register_fct_meta ( "debyecv", 1, "(x): Debye heat capacity" );
    G->register_fct_d_d  ( "debyecv", func_debye_cv );
    G->register_fct_meta ( "erfP", 1, "(x): ?" );
    G->register_fct_d_d  ( "erfP", func_erfP );
    G->register_fct_meta ( "erfQ", 1, "(x): ?" );
    G->register_fct_d_d  ( "erfQ", func_erfQ );
    G->register_fct_meta ( "erf", 1, "(x): error function of x" );
    G->register_fct_d_d  ( "erf", func_erf );
    G->register_fct_meta ( "erfc", 1, "(x): complementary error function of x" );
    G->register_fct_d_d  ( "erfc", func_erfc );
    G->register_fct_meta ( "erfi", 1, "(x): imaginary error function of x" );
    G->register_fct_d_d  ( "erfi", func_erfi );
    G->register_fct_meta ( "dawson", 1, "(x): Dawson function of x" );
    G->register_fct_d_d  ( "dawson", func_dawson );
    G->register_fct_meta ( "sinc", 1, "(x): sinus cardinalis, sin(x)/x" );
    G->register_fct_d_d  ( "sinc", func_sinc );
    G->register_fct_meta ( "ceil", 1, "(x): the smallest integer i with i>=x" );
    G->register_fct_i_d  ( "ceil", func_ceil );
    G->register_fct_meta ( "floor", 1, "(x): the largest integer i with i<=x" );
    G->register_fct_i_d  ( "floor", func_floor );
    G->register_fct_meta ( "nint", 1, "(x): the integer nearest to x" );
    G->register_fct_i_d  ( "nint", func_nint );

    G->register_fct_meta ( "diehl", 1, "(cauchywid/gausswid): normalized convolution gauss(*)cauchy" );
    G->register_fct_d_d  ( "diehl", func_diehl );
    G->register_fct_meta ( "lndiehl", 1, "(cauchywid/gausswid): ln of normalized convolution gauss(*)cauchy" );
    G->register_fct_d_d  ( "lndiehl", func_lndiehl );

    // f(2 args)
    G->register_fct_meta ( "throw_if", 2, "(x,s): throw s if x" );
    G->register_fct_i_is ( "throw_if", func_throw_if );
    G->register_fct_meta ( "throw_unless", 1, "(x,s): throw s if !x" );
    G->register_fct_i_is ( "throw_unless", func_throw_unless );
    G->register_fct_meta ( "min2", 2, "(x,y): the smaller of the two arguments x and y" );
    G->register_fct_i_ii ( "min2", func_min );
    G->register_fct_d_dd ( "min2", func_min );
    G->register_fct_meta ( "max2", 2, "(x,y): the smaller of the two arguments x and y" );
    G->register_fct_i_ii ( "max2", func_max );
    G->register_fct_d_dd ( "max2", func_max );
    G->register_fct_meta ( "ran", 2, "(x,y): a random number between x and y" );
    G->register_fct_d_dd ( "ran", func_ran );

    G->register_fct_meta ( "gauss", 2, "(x,s): the normalized Gaussian exp(-x^2/2/s^2)/sqrt(2 pi)/s" );
    G->register_fct_d_dd ( "gauss", func_gauss );
    G->register_fct_meta ( "gnn", 2, "(x,s): the unnormalized Gaussian exp(-x^2/2/s^2)" );
    G->register_fct_d_dd ( "gnn", func_gaussnn );
    G->register_fct_meta ( "cauchy", 2, "(x,w): the Cauchy-Lorentz function ?/(x^2+w^2)" );
    G->register_fct_d_dd ( "cauchy", func_cauchy );

    // f(3 args)
    G->register_fct_meta ( "rehavneg", 3, "(x,y,z): real part of the Havriliak-Negami function" );
    G->register_fct_d_ddd( "rehavneg", func_re_havneg );
    G->register_fct_meta ( "imhavneg", 3, "(x,y,z): imaginary part of the Havriliak-Negami function" );
    G->register_fct_d_ddd( "imhavneg", func_im_havneg );
    G->register_fct_meta ( "q4w", 3, "(x,y,z): ?" );
    G->register_fct_d_ddd( "q4w", func_q4w );
    G->register_fct_meta ( "cauchy2", 3, "(x,y,z): ?" );
    G->register_fct_d_ddd( "cauchy2", func_cauchy2 );
    G->register_fct_meta ( "kwwc", 3, "(w,tau,b): Fourier cosine transform (t->w) of exp((t/tau)^b)" );
    G->register_fct_d_ddd( "kwwc", func_kwwc );
    G->register_fct_meta ( "kwws", 3, "(w,tau,b): Fourier sine transform (t->w) of exp((t/tau)^b)" );
    G->register_fct_d_ddd( "kwws", func_kwws );
    G->register_fct_meta ( "kwwp", 3, "(w,tau,b): primitive of Fourier cosine transform (t->w) of exp((t/tau)^b)" );
    G->register_fct_d_ddd( "kwwp", func_kwwp );
    G->register_fct_meta ( "voigt", 3, "(x,sigma,gamma): convolution of Gaussian(x,sigma) and Lorentzian(x,gamma)" );
    G->register_fct_d_ddd( "voigt", func_voigt );
    G->register_fct_meta ( "zorn", 3, "(I,<I>,s): Zorn's multiple-scattering corrected elastic intensity" );
    G->register_fct_d_ddd( "zorn", func_zorn );
    G->register_fct_meta ( "zorn2", 3, "(q,<u^2>,s): Zorn's multiple-scattering corrected Gaussian elastic intensity for Si111" );
    G->register_fct_d_ddd( "zorn2", func_zorn_gauss );
    G->register_fct_meta ( "rrdm", 3, "(w*t0,EA_mean/T,EA_stdv/EA_mean: rotational rate distribution model" );
    G->register_fct_d_ddd( "rrdm", func_rrdm );
    G->register_fct_meta ( "ornuhl", 3, "(w,a,b) Ornstein-Uhlenbeck spectrum FT[exp(-|b|t-a(1-exp(-t)))]" );
    G->register_fct_d_ddd( "ornuhl", func_ornuhl );
    G->register_fct_meta ( "rotdiff", 3, "(w,tau,qb: rotational diffusion spectrum)" );
    G->register_fct_d_ddd( "rotdiff", func_rotdiff );
}
