//**************************************************************************************************
//*  FRIDA: fast reliable interactive data analysis
//*  (C) Joachim Wuttke 1990-, v2(C++) 2001-
//*  http://apps.jcns.fz-juelich.de/frida
//**************************************************************************************************

//! \file  olf.hpp
//! \brief COlo, COld, COld online files

#include "ptr.hpp"
#include "coord.hpp"


//! Online object. Virtual base class for COld, COlc.

class COlo {
 public:
    // Data records:
    vector<PSlice> V;
    string name;
    vector<string> lDoc;
    CCoord xco, yco;
    vector<CCoord> ZCo;
    vector<CParam> RPar;

    // Online state records:
    bool as_on_disk;

    COlo() : as_on_disk(false) {}
    virtual ~COlo() {} ///< magic: makes dynamic_cast possible

    void clear() { V.clear(); }
    void remove( int j ) { V.erase( V.begin()+j ); }
    void remove_z( int iz );
    string info_line( int j ) const;

    virtual void copy_mainvec( ROlo fin ) =0;

    int nJ() const { return V.size(); }
    int nZ() const { return ZCo.size(); }
    RObj z( int j, int iz ) const;
    virtual PSlice new_slice( int j ) const = 0;
    virtual POlo new_olo( bool modified=true ) const = 0;
    virtual CCoord coord( class CVar* var ) const = 0;
    virtual void set_coord( class CVar* var, CCoord& co ) = 0;
};


//! Online data file. This is Frida's data model.

class COld : public COlo {
 public:
    COld() : COlo() {}
    explicit COld( class COlc const* fc );

    // Trivially duplicated functions:
    PSpec VS( int j ) const;
    PSlice new_slice( int j ) const;
    POlo new_olo( bool modified=true ) const;
    POld new_old( bool modified=true ) const;
    void copy_mainvec( ROlo fin );

    // Pertinent functions:
    int nPts(int j) const;
    int nPts() const; // 0 unless all spectra j have same nPts(j)
    CCoord coord( class CVar* var ) const;
    void set_coord( class CVar* var, CCoord& co );

    void purge_dy();
};


//! Online curve file.

class COlc : public COlo {
 private:
    RObjVec eval_curve_expr( const vector<double>& vt, int k, int j, bool want_error ) const;
    RObjVec eval_curve_syscall( const vector<double>& vt, int k, int j, bool want_error ) const;

public:
    // Function:
    enum TEva    { _EXPR, _SCR, _SCR_INTP } evaMode; ///< How curve is defined
    enum TScrInp { _STDIN, _CMDLIN } scrInpMode; ///< Where pars are read from
    string expr;                ///< The curve's definition (input expression).
    RNode T;                    ///< The curve's definition (parsed tree).
    int nP;                     ///< Number of parameters.

    // Refined description:
    vector<CCoord> PCo;         ///< Names and units of fit parameters.

    // Fitting:
    int           kd;           ///< Index of data file to be fitted.
    int           kconv;        ///< Index of resolution file to convolve with.
    enum TWgt { _LIN, _LOG, _VAR, _VARC, _VARD }
                  weighing;     ///< Weight mode for fitting.
    bool          plot_to_grid; ///< Use grid from file 'kd' ?
    string        range_expr;   ///< Restricts points to be fitted.
    RNode         range_T;      ///< Parsed range_expr.
    double        chi2;         ///< Global chi^2

    COlc() :
        COlo(), evaMode(_EXPR), scrInpMode(_STDIN),
        kd(-1), kconv(-1), weighing( _VAR ), plot_to_grid( false ),
        range_expr(""), range_T(NULL), chi2(0) {};

    // Trivially duplicated functions:
    PCurve VC( int j ) const;
    PSlice new_slice( int j ) const;
    POlo new_olo( bool modified=true ) const;
    POlc new_olc( bool modified=true ) const;
    void copy_mainvec( ROlo fin );

    // Pertinent functions:
    bool has_global_par() const;
    string weight_str() const;
    vector<string> info_settings() const;
    string info_table_header() const;
    void print_info_table() const;
    CCoord coord( class CVar* var ) const;
    void set_coord( class CVar* var, CCoord& co );

    // The following nontrivial function are implemented in curve.cpp:
    void curve_set_defaults();
    void clear_fits_outcome();
    void parse_function( const string& expr );
    void curve_query( const string& quest );
    RObjVec eval_curve( const vector<double>& vt, int k, int j, bool want_error ) const;
    RObj eval_curve_scalar( double arg, int k, int j, bool want_error ) const;
};


//! Casts.

#define P2D(p) boost::dynamic_pointer_cast<      class COld>(p)
#define R2D(p) boost::dynamic_pointer_cast<const class COld>(p)
#define P2C(p) boost::dynamic_pointer_cast<      class COlc>(p)
#define R2C(p) boost::dynamic_pointer_cast<const class COlc>(p)
